/* Extracted from anet.c to work properly with Hiredict error reporting.
 *
 * Copyright (c) 2009-2011, Salvatore Sanfilippo <antirez at gmail dot com>
 * Copyright (c) 2010-2014, Pieter Noordhuis <pcnoordhuis at gmail dot com>
 * Copyright (c) 2015, Matt Stancliff <matt at genges dot com>,
 *                     Jan-Erik Rediger <janerik at fnordig dot com>
 *
 * SPDX-FileCopyrightText: 2024 Hiredict Contributors
 * SPDX-FileCopyrightText: 2024 Salvatore Sanfilippo <antirez at gmail dot com>
 * SPDX-FileCopyrightText: 2024 Pieter Noordhuis <pcnoordhuis at gmail dot com>
 * SPDX-FileCopyrightText: 2024 Matt Stancliff <matt at genges dot com>
 * SPDX-FileCopyrightText: 2024 Jan-Erik Rediger <janerik at fnordig dot com>
 *
 * SPDX-License-Identifier: BSD-3-Clause
 * SPDX-License-Identifier: LGPL-3.0-or-later
 *
 */

#include "fmacros.h"
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>

#include "net.h"
#include "sds.h"
#include "sockcompat.h"
#include "win32.h"

/* Defined in hiredict.c */
void __redictSetError(redictContext *c, int type, const char *str);

int redictContextUpdateCommandTimeout(redictContext *c, const struct timeval *timeout);

void redictNetClose(redictContext *c) {
    if (c && c->fd != REDICT_INVALID_FD) {
        close(c->fd);
        c->fd = REDICT_INVALID_FD;
    }
}

ssize_t redictNetRead(redictContext *c, char *buf, size_t bufcap) {
    ssize_t nread = recv(c->fd, buf, bufcap, 0);
    if (nread == -1) {
        if ((errno == EWOULDBLOCK && !(c->flags & REDICT_BLOCK)) || (errno == EINTR)) {
            /* Try again later */
            return 0;
        } else if(errno == ETIMEDOUT && (c->flags & REDICT_BLOCK)) {
            /* especially in windows */
            __redictSetError(c, REDICT_ERR_TIMEOUT, "recv timeout");
            return -1;
        } else {
            __redictSetError(c, REDICT_ERR_IO, strerror(errno));
            return -1;
        }
    } else if (nread == 0) {
        __redictSetError(c, REDICT_ERR_EOF, "Server closed the connection");
        return -1;
    } else {
        return nread;
    }
}

ssize_t redictNetWrite(redictContext *c) {
    ssize_t nwritten;

    nwritten = send(c->fd, c->obuf, sdslen(c->obuf), 0);
    if (nwritten < 0) {
        if ((errno == EWOULDBLOCK && !(c->flags & REDICT_BLOCK)) || (errno == EINTR)) {
            /* Try again */
            return 0;
        } else {
            __redictSetError(c, REDICT_ERR_IO, strerror(errno));
            return -1;
        }
    }

    return nwritten;
}

static void __redictSetErrorFromErrno(redictContext *c, int type, const char *prefix) {
    int errorno = errno;  /* snprintf() may change errno */
    char buf[128] = { 0 };
    size_t len = 0;

    if (prefix != NULL)
        len = snprintf(buf,sizeof(buf),"%s: ",prefix);
    strerror_r(errorno, (char *)(buf + len), sizeof(buf) - len);
    __redictSetError(c,type,buf);
}

static int redictSetReuseAddr(redictContext *c) {
    int on = 1;
    if (setsockopt(c->fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,NULL);
        redictNetClose(c);
        return REDICT_ERR;
    }
    return REDICT_OK;
}

static int redictCreateSocket(redictContext *c, int type) {
    redictFD s;
    if ((s = socket(type, SOCK_STREAM, 0)) == REDICT_INVALID_FD) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,NULL);
        return REDICT_ERR;
    }
    c->fd = s;
    if (type == AF_INET) {
        if (redictSetReuseAddr(c) == REDICT_ERR) {
            return REDICT_ERR;
        }
    }
    return REDICT_OK;
}

static int redictSetBlocking(redictContext *c, int blocking) {
#ifndef _WIN32
    int flags;

    /* Set the socket nonblocking.
     * Note that fcntl(2) for F_GETFL and F_SETFL can't be
     * interrupted by a signal. */
    if ((flags = fcntl(c->fd, F_GETFL)) == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,"fcntl(F_GETFL)");
        redictNetClose(c);
        return REDICT_ERR;
    }

    if (blocking)
        flags &= ~O_NONBLOCK;
    else
        flags |= O_NONBLOCK;

    if (fcntl(c->fd, F_SETFL, flags) == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,"fcntl(F_SETFL)");
        redictNetClose(c);
        return REDICT_ERR;
    }
#else
    u_long mode = blocking ? 0 : 1;
    if (ioctl(c->fd, FIONBIO, &mode) == -1) {
        __redictSetErrorFromErrno(c, REDICT_ERR_IO, "ioctl(FIONBIO)");
        redictNetClose(c);
        return REDICT_ERR;
    }
#endif /* _WIN32 */
    return REDICT_OK;
}

int redictKeepAlive(redictContext *c, int interval) {
    int val = 1;
    redictFD fd = c->fd;

    /* TCP_KEEPALIVE makes no sense with AF_UNIX connections */
    if (c->connection_type == REDICT_CONN_UNIX)
        return REDICT_ERR;

#ifndef _WIN32
    if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &val, sizeof(val)) == -1){
        __redictSetError(c,REDICT_ERR_OTHER,strerror(errno));
        return REDICT_ERR;
    }

    val = interval;

#if defined(__APPLE__) && defined(__MACH__)
    if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPALIVE, &val, sizeof(val)) < 0) {
        __redictSetError(c,REDICT_ERR_OTHER,strerror(errno));
        return REDICT_ERR;
    }
#else
#if defined(__GLIBC__) && !defined(__FreeBSD_kernel__)
    if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPIDLE, &val, sizeof(val)) < 0) {
        __redictSetError(c,REDICT_ERR_OTHER,strerror(errno));
        return REDICT_ERR;
    }

    val = interval/3;
    if (val == 0) val = 1;
    if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPINTVL, &val, sizeof(val)) < 0) {
        __redictSetError(c,REDICT_ERR_OTHER,strerror(errno));
        return REDICT_ERR;
    }

    val = 3;
    if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPCNT, &val, sizeof(val)) < 0) {
        __redictSetError(c,REDICT_ERR_OTHER,strerror(errno));
        return REDICT_ERR;
    }
#endif
#endif
#else
    int res;

    res = win32_redictKeepAlive(fd, interval * 1000);
    if (res != 0) {
        __redictSetError(c, REDICT_ERR_OTHER, strerror(res));
        return REDICT_ERR;
    }
#endif
    return REDICT_OK;
}

int redictSetTcpNoDelay(redictContext *c) {
    int yes = 1;
    if (setsockopt(c->fd, IPPROTO_TCP, TCP_NODELAY, &yes, sizeof(yes)) == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,"setsockopt(TCP_NODELAY)");
        redictNetClose(c);
        return REDICT_ERR;
    }
    return REDICT_OK;
}

int redictContextSetTcpUserTimeout(redictContext *c, unsigned int timeout) {
    int res;
#ifdef TCP_USER_TIMEOUT
    res = setsockopt(c->fd, IPPROTO_TCP, TCP_USER_TIMEOUT, &timeout, sizeof(timeout));
#else
    res = -1;
    errno = ENOTSUP;
    (void)timeout;
#endif
    if (res == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,"setsockopt(TCP_USER_TIMEOUT)");
        redictNetClose(c);
        return REDICT_ERR;
    }
    return REDICT_OK;
}

#define __MAX_MSEC (((LONG_MAX) - 999) / 1000)

static int redictContextTimeoutMsec(redictContext *c, long *result)
{
    const struct timeval *timeout = c->connect_timeout;
    long msec = -1;

    /* Only use timeout when not NULL. */
    if (timeout != NULL) {
        if (timeout->tv_usec > 1000000 || timeout->tv_sec > __MAX_MSEC) {
            __redictSetError(c, REDICT_ERR_IO, "Invalid timeout specified");
            *result = msec;
            return REDICT_ERR;
        }

        msec = (timeout->tv_sec * 1000) + ((timeout->tv_usec + 999) / 1000);

        if (msec < 0 || msec > INT_MAX) {
            msec = INT_MAX;
        }
    }

    *result = msec;
    return REDICT_OK;
}

static long redictPollMillis(void) {
#ifndef _MSC_VER
    struct timespec now;
    clock_gettime(CLOCK_MONOTONIC, &now);
    return (now.tv_sec * 1000) + now.tv_nsec / 1000000;
#else
    FILETIME ft;
    GetSystemTimeAsFileTime(&ft);
    return (((long long)ft.dwHighDateTime << 32) | ft.dwLowDateTime) / 10;
#endif
}

static int redictContextWaitReady(redictContext *c, long msec) {
    struct pollfd wfd;
    long end;
    int res;

    if (errno != EINPROGRESS) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,NULL);
        redictNetClose(c);
        return REDICT_ERR;
    }

    wfd.fd = c->fd;
    wfd.events = POLLOUT;
    end = msec >= 0 ? redictPollMillis() + msec : 0;

    while ((res = poll(&wfd, 1, msec)) <= 0) {
        if (res < 0 && errno != EINTR) {
            __redictSetErrorFromErrno(c, REDICT_ERR_IO, "poll(2)");
            redictNetClose(c);
            return REDICT_ERR;
        } else if (res == 0 || (msec >= 0 && redictPollMillis() >= end)) {
            errno = ETIMEDOUT;
            __redictSetErrorFromErrno(c, REDICT_ERR_IO, NULL);
            redictNetClose(c);
            return REDICT_ERR;
        } else {
            /* res < 0 && errno == EINTR, try again */
        }
    }

    if (redictCheckConnectDone(c, &res) != REDICT_OK || res == 0) {
        redictCheckSocketError(c);
        return REDICT_ERR;
    }

    return REDICT_OK;
}

int redictCheckConnectDone(redictContext *c, int *completed) {
    int rc = connect(c->fd, (const struct sockaddr *)c->saddr, c->addrlen);
    if (rc == 0) {
        *completed = 1;
        return REDICT_OK;
    }
    int error = errno;
    if (error == EINPROGRESS) {
        /* must check error to see if connect failed.  Get the socket error */
        int fail, so_error;
        socklen_t optlen = sizeof(so_error);
        fail = getsockopt(c->fd, SOL_SOCKET, SO_ERROR, &so_error, &optlen);
        if (fail == 0) {
            if (so_error == 0) {
                /* Socket is connected! */
                *completed = 1;
                return REDICT_OK;
            }
            /* connection error; */
            errno = so_error;
            error = so_error;
        }
    }
    switch (error) {
    case EISCONN:
        *completed = 1;
        return REDICT_OK;
    case EALREADY:
    case EWOULDBLOCK:
        *completed = 0;
        return REDICT_OK;
    default:
        return REDICT_ERR;
    }
}

int redictCheckSocketError(redictContext *c) {
    int err = 0, errno_saved = errno;
    socklen_t errlen = sizeof(err);

    if (getsockopt(c->fd, SOL_SOCKET, SO_ERROR, &err, &errlen) == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,"getsockopt(SO_ERROR)");
        return REDICT_ERR;
    }

    if (err == 0) {
        err = errno_saved;
    }

    if (err) {
        errno = err;
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,NULL);
        return REDICT_ERR;
    }

    return REDICT_OK;
}

int redictContextSetTimeout(redictContext *c, const struct timeval tv) {
    const void *to_ptr = &tv;
    size_t to_sz = sizeof(tv);

    if (redictContextUpdateCommandTimeout(c, &tv) != REDICT_OK) {
        __redictSetError(c, REDICT_ERR_OOM, "Out of memory");
        return REDICT_ERR;
    }
    if (setsockopt(c->fd,SOL_SOCKET,SO_RCVTIMEO,to_ptr,to_sz) == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,"setsockopt(SO_RCVTIMEO)");
        return REDICT_ERR;
    }
    if (setsockopt(c->fd,SOL_SOCKET,SO_SNDTIMEO,to_ptr,to_sz) == -1) {
        __redictSetErrorFromErrno(c,REDICT_ERR_IO,"setsockopt(SO_SNDTIMEO)");
        return REDICT_ERR;
    }
    return REDICT_OK;
}

int redictContextUpdateConnectTimeout(redictContext *c, const struct timeval *timeout) {
    /* Same timeval struct, short circuit */
    if (c->connect_timeout == timeout)
        return REDICT_OK;

    /* Allocate context timeval if we need to */
    if (c->connect_timeout == NULL) {
        c->connect_timeout = hi_malloc(sizeof(*c->connect_timeout));
        if (c->connect_timeout == NULL)
            return REDICT_ERR;
    }

    memcpy(c->connect_timeout, timeout, sizeof(*c->connect_timeout));
    return REDICT_OK;
}

int redictContextUpdateCommandTimeout(redictContext *c, const struct timeval *timeout) {
    /* Same timeval struct, short circuit */
    if (c->command_timeout == timeout)
        return REDICT_OK;

    /* Allocate context timeval if we need to */
    if (c->command_timeout == NULL) {
        c->command_timeout = hi_malloc(sizeof(*c->command_timeout));
        if (c->command_timeout == NULL)
            return REDICT_ERR;
    }

    memcpy(c->command_timeout, timeout, sizeof(*c->command_timeout));
    return REDICT_OK;
}

static int _redictContextConnectTcp(redictContext *c, const char *addr, int port,
                                   const struct timeval *timeout,
                                   const char *source_addr) {
    redictFD s;
    int rv, n;
    char _port[6];  /* strlen("65535"); */
    struct addrinfo hints, *servinfo, *bservinfo, *p, *b;
    int blocking = (c->flags & REDICT_BLOCK);
    int reuseaddr = (c->flags & REDICT_REUSEADDR);
    int reuses = 0;
    long timeout_msec = -1;

    servinfo = NULL;
    c->connection_type = REDICT_CONN_TCP;
    c->tcp.port = port;

    /* We need to take possession of the passed parameters
     * to make them reusable for a reconnect.
     * We also carefully check we don't free data we already own,
     * as in the case of the reconnect method.
     *
     * This is a bit ugly, but atleast it works and doesn't leak memory.
     **/
    if (c->tcp.host != addr) {
        hi_free(c->tcp.host);

        c->tcp.host = hi_strdup(addr);
        if (c->tcp.host == NULL)
            goto oom;
    }

    if (timeout) {
        if (redictContextUpdateConnectTimeout(c, timeout) == REDICT_ERR)
            goto oom;
    } else {
        hi_free(c->connect_timeout);
        c->connect_timeout = NULL;
    }

    if (redictContextTimeoutMsec(c, &timeout_msec) != REDICT_OK) {
        goto error;
    }

    if (source_addr == NULL) {
        hi_free(c->tcp.source_addr);
        c->tcp.source_addr = NULL;
    } else if (c->tcp.source_addr != source_addr) {
        hi_free(c->tcp.source_addr);
        c->tcp.source_addr = hi_strdup(source_addr);
    }

    snprintf(_port, 6, "%d", port);
    memset(&hints,0,sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    /* DNS lookup. To use dual stack, set both flags to prefer both IPv4 and
     * IPv6. By default, for historical reasons, we try IPv4 first and then we
     * try IPv6 only if no IPv4 address was found. */
    if (c->flags & REDICT_PREFER_IPV6 && c->flags & REDICT_PREFER_IPV4)
        hints.ai_family = AF_UNSPEC;
    else if (c->flags & REDICT_PREFER_IPV6)
        hints.ai_family = AF_INET6;
    else
        hints.ai_family = AF_INET;

    rv = getaddrinfo(c->tcp.host, _port, &hints, &servinfo);
    if (rv != 0 && hints.ai_family != AF_UNSPEC) {
        /* Try again with the other IP version. */
        hints.ai_family = (hints.ai_family == AF_INET) ? AF_INET6 : AF_INET;
        rv = getaddrinfo(c->tcp.host, _port, &hints, &servinfo);
    }
    if (rv != 0) {
        __redictSetError(c, REDICT_ERR_OTHER, gai_strerror(rv));
        return REDICT_ERR;
    }
    for (p = servinfo; p != NULL; p = p->ai_next) {
addrretry:
        if ((s = socket(p->ai_family,p->ai_socktype,p->ai_protocol)) == REDICT_INVALID_FD)
            continue;

        c->fd = s;
        if (redictSetBlocking(c,0) != REDICT_OK)
            goto error;
        if (c->tcp.source_addr) {
            int bound = 0;
            /* Using getaddrinfo saves us from self-determining IPv4 vs IPv6 */
            if ((rv = getaddrinfo(c->tcp.source_addr, NULL, &hints, &bservinfo)) != 0) {
                char buf[128];
                snprintf(buf,sizeof(buf),"Can't get addr: %s",gai_strerror(rv));
                __redictSetError(c,REDICT_ERR_OTHER,buf);
                goto error;
            }

            if (reuseaddr) {
                n = 1;
                if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*) &n,
                               sizeof(n)) < 0) {
                    freeaddrinfo(bservinfo);
                    goto error;
                }
            }

            for (b = bservinfo; b != NULL; b = b->ai_next) {
                if (bind(s,b->ai_addr,b->ai_addrlen) != -1) {
                    bound = 1;
                    break;
                }
            }
            freeaddrinfo(bservinfo);
            if (!bound) {
                char buf[128];
                snprintf(buf,sizeof(buf),"Can't bind socket: %s",strerror(errno));
                __redictSetError(c,REDICT_ERR_OTHER,buf);
                goto error;
            }
        }

        /* For repeat connection */
        hi_free(c->saddr);
        c->saddr = hi_malloc(p->ai_addrlen);
        if (c->saddr == NULL)
            goto oom;

        memcpy(c->saddr, p->ai_addr, p->ai_addrlen);
        c->addrlen = p->ai_addrlen;

        if (connect(s,p->ai_addr,p->ai_addrlen) == -1) {
            if (errno == EHOSTUNREACH) {
                redictNetClose(c);
                continue;
            } else if (errno == EINPROGRESS) {
                if (blocking) {
                    goto wait_for_ready;
                }
                /* This is ok.
                 * Note that even when it's in blocking mode, we unset blocking
                 * for `connect()`
                 */
            } else if (errno == EADDRNOTAVAIL && reuseaddr) {
                if (++reuses >= REDICT_CONNECT_RETRIES) {
                    goto error;
                } else {
                    redictNetClose(c);
                    goto addrretry;
                }
            } else {
                wait_for_ready:
                if (redictContextWaitReady(c,timeout_msec) != REDICT_OK)
                    goto error;
                if (redictSetTcpNoDelay(c) != REDICT_OK)
                    goto error;
            }
        }
        if (blocking && redictSetBlocking(c,1) != REDICT_OK)
            goto error;

        c->flags |= REDICT_CONNECTED;
        rv = REDICT_OK;
        goto end;
    }
    if (p == NULL) {
        char buf[128];
        snprintf(buf,sizeof(buf),"Can't create socket: %s",strerror(errno));
        __redictSetError(c,REDICT_ERR_OTHER,buf);
        goto error;
    }

oom:
    __redictSetError(c, REDICT_ERR_OOM, "Out of memory");
error:
    rv = REDICT_ERR;
end:
    if(servinfo) {
        freeaddrinfo(servinfo);
    }

    return rv;  // Need to return REDICT_OK if alright
}

int redictContextConnectTcp(redictContext *c, const char *addr, int port,
                           const struct timeval *timeout) {
    return _redictContextConnectTcp(c, addr, port, timeout, NULL);
}

int redictContextConnectBindTcp(redictContext *c, const char *addr, int port,
                               const struct timeval *timeout,
                               const char *source_addr) {
    return _redictContextConnectTcp(c, addr, port, timeout, source_addr);
}

int redictContextConnectUnix(redictContext *c, const char *path, const struct timeval *timeout) {
#ifndef _WIN32
    int blocking = (c->flags & REDICT_BLOCK);
    struct sockaddr_un *sa;
    long timeout_msec = -1;

    if (redictCreateSocket(c,AF_UNIX) < 0)
        return REDICT_ERR;
    if (redictSetBlocking(c,0) != REDICT_OK)
        return REDICT_ERR;

    c->connection_type = REDICT_CONN_UNIX;
    if (c->unix_sock.path != path) {
        hi_free(c->unix_sock.path);

        c->unix_sock.path = hi_strdup(path);
        if (c->unix_sock.path == NULL)
            goto oom;
    }

    if (timeout) {
        if (redictContextUpdateConnectTimeout(c, timeout) == REDICT_ERR)
            goto oom;
    } else {
        hi_free(c->connect_timeout);
        c->connect_timeout = NULL;
    }

    if (redictContextTimeoutMsec(c,&timeout_msec) != REDICT_OK)
        return REDICT_ERR;

    /* Don't leak sockaddr if we're reconnecting */
    if (c->saddr) hi_free(c->saddr);

    sa = (struct sockaddr_un*)(c->saddr = hi_malloc(sizeof(struct sockaddr_un)));
    if (sa == NULL)
        goto oom;

    c->addrlen = sizeof(struct sockaddr_un);
    sa->sun_family = AF_UNIX;
    strncpy(sa->sun_path, path, sizeof(sa->sun_path) - 1);
    if (connect(c->fd, (struct sockaddr*)sa, sizeof(*sa)) == -1) {
        if (errno == EINPROGRESS && !blocking) {
            /* This is ok. */
        } else {
            if (redictContextWaitReady(c,timeout_msec) != REDICT_OK)
                return REDICT_ERR;
        }
    }

    /* Reset socket to be blocking after connect(2). */
    if (blocking && redictSetBlocking(c,1) != REDICT_OK)
        return REDICT_ERR;

    c->flags |= REDICT_CONNECTED;
    return REDICT_OK;
#else
    /* We currently do not support Unix sockets for Windows. */
    /* TODO(m): https://devblogs.microsoft.com/commandline/af_unix-comes-to-windows/ */
    errno = EPROTONOSUPPORT;
    return REDICT_ERR;
#endif /* _WIN32 */
oom:
    __redictSetError(c, REDICT_ERR_OOM, "Out of memory");
    return REDICT_ERR;
}
